
package RRDutils;

use warnings;
use strict;

use RRDs;

BEGIN {
   use Exporter ();
    our ( $VERSION , @ISA , @EXPORT , @EXPORT_OK , %EXPORT_TAGS );

   $VERSION="5.0.1";
   @ISA = qw(Exporter);   
   @EXPORT = qw(
      changeDefs $confdir $datadir $graphdir
      getHosts getTags %getNAME
      );
   @EXPORT_OK = qw(
      getStep getStart getDSs getRRAs
      @figSize %graphInterval @colors
      getIntervals getGraphs getLabels
      rrdCreate rrdGraph
      openWeb closeWeb htmlAnchor
      %getDEFs %getNAME
      );
   %EXPORT_TAGS = (
      Create => [ qw(getStep getStart getDSs getRRAs rrdCreate) ]
    , Graph => [ qw(
        @figSize %graphInterval @colors
        getIntervals getGraphs getLabels rrdGraph
        %getDEFs %getNAME
        ) ]
    , Web => [ qw(getIntervals getGraphs
        openWeb closeWeb htmlAnchor) ]
      );
   }

our ( $confdir , $datadir , $graphdir );
our ( %getDEFs , %getNAME );

my $rootdir = "/var/lib/rrd";
$rootdir = $ENV{"ProgramFiles"}."\\rrdUtils" if $ENV{"ProgramFiles"};

$confdir = "$rootdir/conf";
$datadir = "$rootdir/data";
$graphdir = "$rootdir/graphs";

my $defStep = 300;
my @defGraph = ( 300, 90 );

our ( @figSize , %graphInterval , @colors );

@figSize = ( "--width" , "$defGraph[0]" , "--height" , "$defGraph[1]" );

%graphInterval = (
   'daily' => -86400 ,
   'weekly' => -604800 ,
   'monthly' => -2592000 ,
   'quarterly' => -10368000 ,
   'yearly' => -41472000 );

@colors= qw(ff0000 00ff00 0000ff aa0000 00aa00 0000aa
   880000 008800 000088 660000 006600 000066
   ff0000 00ff00 0000ff ff0000 00ff00 0000ff
   ff0000 00ff00 0000ff ff0000 00ff00 0000ff
   ff0000 00ff00 0000ff ff0000 00ff00 0000ff
   ff0000 00ff00 0000ff ff0000 00ff00 0000ff
   ff0000 00ff00 0000ff ff0000 00ff00 0000ff
   ff0000 00ff00 0000ff ff0000 00ff00 0000ff);


sub changeDefs($) {
   my $conf = shift;
   return unless open CFG , "$confdir/$conf";
   while (<CFG>) {
      next if /^\s*#/; s/\s*$//;
      if ( /^DataDir\s+(.+)$/ ) {
         $datadir=$1;
         }
      if ( /^GraphDir\s(.+)$/ ) {
         $graphdir=$1;
         }
      if ( /^DefGraph\s+([0-9]+)\s+([0-9]+)$/ ) {
         @defGraph = ( $1 , $2 );
         @figSize = ( "--width" , "$defGraph[0]" , "--height" , "$defGraph[1]" );
         }
      }
   close CFG;
   }

sub getStep($) {
   my $conf = shift;
   my $step;
   return unless open CFG , "$confdir/$conf";
   while (<CFG>) {
      next if /^\s*#/; s/\s*$//;
      if ( /^BaseStep\s+([0-9]*)/ ) {
         warn "WARNING : base step redefined\n" if $step;
         $step=$1;
         }
      }
   close CFG;
   $step = $defStep unless $step;
   return $step;
   }

sub getStart($) {
   my $conf = shift;
   my $start;
   return unless open CFG , "$confdir/$conf";
   while (<CFG>) {
      next if /^\s*#/; s/\s*$//;
      if ( /^Start\s+(.*)/ ) {
         warn "WARNING : Start time for RRDB defined twice" if $start;
         $start=$1;
         }
      }
   return $start;
   }

sub getHosts($) {
   my ( $conf , $tag ) = split /\./ , shift ;
   $tag = "" unless $tag;
   my @hosts;
   return unless open CFG , "$confdir/$conf";
   while (<CFG>) {
      next if /^\s*#/; s/\s*$//;
      if ( s/^H$tag\s+// ) {
         push @hosts , split;
         }
      if ( s/^T\s+// ) {
         push @hosts , [ split ];
         }
      }
   close CFG;
   return @hosts;
   }

sub getTags($) {
   my $conf = shift;
   my %tags;
   return unless open CFG , "$confdir/$conf";
   while (<CFG>) {
      next if /^\s*#/; s/\s*$//;
      if ( s/^H([^\s]+)\s// ) {
         $tags{$1} = $1;
         }
      }
   close CFG;
   return keys %tags;
   }


sub getDSs($) {
   my $conf = shift;
   my %types = (
      'G' => 'GAUGE' ,
      'C' => 'COUNTER' ,
      'D' => 'DERIVE' ,
      'A' => 'ABSOLUTE' );
   my ( @DSs , $label );
   my $step = getStep $conf;
   return unless open CFG , "$confdir/$conf";
   $_ = <CFG>; s/\s*$//;
   while ( 1 ) {
      last if eof CFG;
      if ( /^[^#]/ && /^([^\s\[]*)?\[([^\s\/]+)\](\s+(.*))?$/ ) {
         $label = $2;
         my ( $type, $beat, $min, $max ) = ( "COUNTER", 3 , "U", "U" );
         $_ = <CFG>; $_ =~ s/\s*$//;
         if ( s/^((C|G|D|A)\s?)?\s*([0-9\.]+)?(\s+((U|-?[0-9\.]+)\s+)?(U|-?[0-9\.]+))?$// ) {
            $type = $types{$2} if $2;
            $beat = $3 if $3;
            $min = $6 if $6;
            $max = $7 if $7;
            unless ( eof CFG ) { $_ = <CFG> ; s/\s*$//; }
            }
         push @DSs , "DS:$label:$type:" . $beat * $step . ":$min:$max";
      } else {
         last if eof CFG;
         $_ = <CFG>; s/\s*$//;
         }
      }
   close CFG;

   return @DSs;
   }

sub getRRAs($) {
   my $conf = shift;
   my @RRAs;
   my $TimeStep = 1;
   return unless open CFG , "$confdir/$conf";
   while ( <CFG> ) {
      next if /^\s*#/; s/\s*$//;
      if ( /^TimeStep\s+(.*)/ ) {
         $TimeStep *= $1;
         next;
         }
      if ( s/^(CF\s+)?(AVERAGE|MAX|MIN|LAST)// ) { #(.*)(\s+([0-9]+))?$/ ) {
         my $CF = $2;
         my ( $xff , $step , $rows ) = ( 0.5 , 1 , 192 );
         /(\s+(0|0?\.[0-9]+)?\s*([1-9][0-9]*)?)?\s+([0-9]+)$/;
         $xff = $2 if $2;
         $step = $3 if $3;
         $rows = $4 if $4;
         push @RRAs , "RRA:$CF:$xff:" . $step * $TimeStep . ":$rows";
         next;
         }
      }
   close CFG;
   @RRAs = ( "RRA:AVERAGE:0.5:1:2048" ) unless @RRAs;
   return @RRAs;
   }


sub getIntervals($) {
   my $conf = shift;
   return unless open CFG , "$confdir/$conf";
   while (<CFG>) {
      next if /^\s*#/; s/\s*$//;
      if ( /^I\s+([^\s]+)\s+(\d+)/ ) {
         $graphInterval{$1} = "-$2";
         }
      }
   close CFG;
   }

sub getSize($) {
   my $size = shift;
   $size =~ /^(.*)x(.*)/;
   my $xsize = $1 || $defGraph[0];
   my $ysize = $2 || $defGraph[1];
   return "--width" , "$xsize" , "--height" , "$ysize";
   }

sub getGraphs($) {
   my $conf = shift;
   my $opts = "";
   my %graphs;
   return unless open CFG , "$confdir/$conf";
   while (<CFG>) {
      next if /^\s*#/; s/\s*$//;
      if ( s/^O(=?)\s+// ) {
         chomp;
         $opts = "$_" if $1;
         $opts .= " $_";
         }
      if ( s/^(P(S|H)?)(_?([a-z_]+))?(\s+([0-9]*x[0-9]*))?\s+(STD|[a-z,]+)\s+// ) {
         my @vars=split;
         my @size = ();
         @size=getSize($6) if $6;
         my $idx = join "." , @vars;
         $idx = $4 if $4;
         $graphs{$idx}{'type'} = $1;
         $graphs{$idx}{'size'} = \@size;
         my $period = $7;
         $period = "daily,weekly,monthly,quarterly,yearly" if $period eq 'STD' ;
         if ( $graphs{$idx}{'periods'} ) {
            $graphs{$idx}{'periods'} .= "," . $period;
         } else {
            $graphs{$idx}{'periods'} = $period;
            }
         $graphs{$idx}{'opts'} = $opts;
         $graphs{$idx}{'ds'} = \@vars;
         }
      }
   close CFG;
   return %graphs;
   }


sub getLabels($) {
   my $conf = shift;
   my %labels;
   return unless open CFG , "$confdir/$conf";
   while (<CFG>) {
      next if /^\s*#/; s/\s*$//;
      if ( /^([^\s\[]*)?\[([^\s\/]+)\](\s+(.*))?$/ ) {
         $labels{$2} = $4 || $1 || $2;
         $labels{$2} =~ s/\s*$//;
         }
      }
   close CFG;
   return %labels;
   }

sub rrdCreate {
   my ( $host , $target , @args ) = @_ ;
   my $rrdname = $getNAME{'rrd'}( $host , $target );
   if ( -e "$rrdname" ) {
      return "$target for $host exists, not overwritten\n";
      }
   mkdir "$datadir/$host";
   mkdir "$graphdir/$host";
   mkdir "$graphdir/$host/$target";
   RRDs::create ( "$rrdname" , @args );
   return RRDs::error;
   }

# To use also with PH plots, the title and other items differ
sub rrdGraph {
   my ( $host , $target , $graph , $period , @graphargs ) = @_ ;
   my $imgname = $getNAME{'img'}( $host , $target , $graph , $period );
   RRDs::graph ( $imgname , "--title=$target @ $host" , @graphargs );
   return RRDs::error;
   }

# FIXME: What about using step for refresh time
my $refresh = "<META HTTP-EQUIV=Refresh CONTENT=3600>" ;
my $style = "bgcolor=#ffffff text=#000000" ;
my $tabstyle = "BORDER=0 CELLPADDING=0 CELLSPACING=10" ;

sub openWeb($$) {
   my ( $file , $title ) = @_ ;
   open my $html , '>' , "$graphdir/$file.html" ;
   print $html "<HTML>\n<HEAD>\n";
   print $html "$refresh\n";
   print $html "<TITLE>$title</TITLE>\n";
   print $html "</HEAD>\n<BODY $style>\n";
   print $html "<TABLE $tabstyle>\n";
   return $html;
   }

sub closeWeb($) {
   my $html = shift ;
   print $html "</TABLE>\n";
   print $html "</BODY>\n</HTML>\n";
   close $html;
   }

sub htmlAnchor($@) {
   my $src = shift ;
   # Thist should be a call to getNAME{'html'}
   my $html = join "/" , @_ ;
   return "<tr><td>" .
      ( $html ? "<A HREF=$html.html>" : "" ) .
      "<IMG SRC=$src>" .
      ( $html ? "</A>" : "" ) .
      "</td></tr>\n";
   }

my %CFs = ( 'a' => 'AVERAGE' , 'm' => 'MIN' , 'M' => 'MAX' , 'l' => 'LAST' );

sub cfType($) {
   my $dsname = shift;
   my $type = 'AVERAGE';
#WinXP   return ( $dsname , $type );
   $dsname =~ s/^(a|m|M|l)_//;
   $type = $CFs{$1} if $1;
   return ( $dsname , $type );
   }

$getDEFs{'PH'} = sub {
   my ( $item , $target , $legend , @hosts ) = @_;
   my ( $n , $cf , @cmd );
   ( $item , $cf ) = cfType $item;
   for $n ( 0 .. $#hosts ) {
      if ( ref($hosts[$n]) eq "ARRAY" ) {
         $hosts[$n] = $hosts[$n]->[0];
         }
      my $rrdname = $getNAME{'rrd'}( $hosts[$n] , $target );
      $rrdname =~ s/:/\\:/;
      push @cmd , "DEF:d$n=$rrdname:$item:$cf";
      push @cmd , "LINE1:d$n#$colors[$n]:$hosts[$n]";
      }
   return @cmd;
   };
$getDEFs{'PS'} = sub {
   my ( $host , $target , $legend , @items ) = @_;
   my ( $item , $cf , @cmd );
   ( $item , $cf ) = cfType $items[0];
   my $rrdname = $getNAME{'rrd'}( $host , $target );
   push @cmd , "DEF:d0=$rrdname:$item:$cf";
   push @cmd , "AREA:d0#$colors[0]:$$legend{$item}";
   my $n;
   for $n ( 1 .. $#items ) {
      my ( $item , $cf ) = cfType $items[$n];
      my $rrdname = $getNAME{'rrd'}( $host , $target );
      $rrdname =~ s/:/\\:/;
      push @cmd , "DEF:d$n=$rrdname:$item:$cf";
      push @cmd , "STACK:d$n#$colors[$n]:$$legend{$item}";
      }
   return @cmd;
   };
$getDEFs{'P'} = sub {
   my ( $host , $target , $legend , @items ) = @_;
   my ( $item , $cf );
   my ( $n , @cmd ) = ( 0 , () );
   for $n ( 0 .. $#items ) {
      ( $item , $cf ) = cfType $items[$n];
      my $rrdname = $getNAME{'rrd'}( $host , $target );
      $rrdname =~ s/:/\\:/;
      push @cmd , "DEF:d$n=$rrdname:$item:$cf";
      push @cmd , "LINE1:d$n#$colors[$n]:$$legend{$item}";
      }
   return @cmd;
   };

# FIXME: no index yet
# IMPROVE: The easiest way is just use a single char as separator, betwen
#          host & target , $graph & period , ....
$getNAME{'rrd'} = sub {
   my ( $host , $target ) = @_ ;
   return "$datadir/$host/$target.rrd";
   };
$getNAME{'img'} = sub {
   my ( $host , $target , $graph , $period ) = @_ ;
   return "$graphdir/$host/$target/$graph.$period.png";
   };
$getNAME{'src'} = sub {
   my $period = pop ;
   return join( "/" , @_ ) . ".$period.png";
   };
sub htmName($$;$) {
   my $host = shift;
   my $target = shift;
   if ( @_ ) { $target .= "/" . shift }
   if ( $host eq "-" ) { return $target }
   return "$graphdir/$host/$target.html";
   }

1;
